/**
 * Copyright 2011 Marco Buzzanca. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright notice, this list of
 *       conditions and the following disclaimer.
 * 
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list
 *       of conditions and the following disclaimer in the documentation and/or other materials
 *       provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are those of the
 * authors and should not be interpreted as representing official policies, either expressed
 * or implied, of the copyright holder.
 */
 
import edu.emory.mathcs.jtransforms.fft.*;
import shx.math.*;
import java.io.*;

public class MotionDeblurFilter
{
	private int height;
	private int width;
	private double[] otf;
	private double snr;
	private DoubleFFT_2D fft;
	
	public MotionDeblurFilter(int width, int height)
	{
		this.width = width;
		this.height = height;
		this.snr = 0.01d;
			
		fft = new DoubleFFT_2D(height, width);
	}
	
	public MotionDeblurFilter(int width, int height, double shift, double angle, double snr)
	{
		this.width = width;
		this.height = height;
		this.snr = snr;
			
		fft = new DoubleFFT_2D(height, width);
		
		calculateMask(shift, angle);
	}
	
	public void apply(double source[])
	{
		if (otf == null) return;
		
		double[] transform = new double[2 * source.length];

		// Fill only the first half of the array, the other half is ignored by realForwardFull()         
        System.arraycopy(source, 0, transform, 0, source.length);
		
		process(transform);
		
		// Discards imaginary part.
		for (int i = 0; i < source.length; i++)
			source[i] = transform[2 * i];
	}
	
	public void apply(float source[])
	{
		if (otf == null) return;
		
		double[] transform = new double[2 * source.length];
		
		// Fill only the first half of the array, the other half is ignored by realForwardFull()
		for (int i = 0; i < source.length; i++)
			transform[i] = (double)source[i];
			
		process(transform);
		
		for (int i = 0; i < source.length; i++)
			source[i] = (float)transform[2 * i];
	}
	
	private void process(double transform[])
	{
		fft.realForwardFull(transform);

		for (int i = 0; i < transform.length; i += 2)
		{
			// transform[i] is the real part of the i-th pixel of the blurred image's transform.
			// transform[i + i] is the imaginary part of the same pixel. Same goes for the OTF.
			Complex blurred = new Complex(transform[i], transform[i + 1]);
			
			// |H|^2 + k is real, no need to store it in a Complex object. It is always non-zero.
			double den = otf[i] * otf[i] + otf[i + 1] * otf[i + 1] + snr;

			// W = H* / (|H|^2 + k)
			Complex weiner = new Complex(otf[i] / den, -otf[i + 1] / den);
			
			// F = W * G
			Complex reconstructed = weiner.mul(blurred);
			
			transform[i] = reconstructed.re();
			transform[i + 1] = reconstructed.im();
		}
		
		fft.complexInverse(transform, true);
	}
	
	// Helper function used in calculateMask()
	private int getShiftedIndex(int x, int y)
	{
		if (x < 0) x += width; 
		if (y < 0) y += height;
				
		return y * width + x;
	}
	
	public double getSNR() { return snr; }
	
	public void setSNR(double snr) { this.snr = snr; }
	
	public double[] getMask() { return otf; }
	
	public void calculateMask(double shift, double angle)
	{
		// Recipent for the OTF information
		otf = new double[2 * width * height];
		
		double length = Math.max(shift, 1d);
		double half = (length - 1d) / 2d;
		
		double phi = Math.toRadians(angle);
		double cosPhi = Math.cos(phi);
		double sinPhi = Math.sin(phi);
		
		int xSign = (cosPhi > 0d ? 1 : -1);
		
		double normFactor = 0d;
		int halfX, halfY;
		
		// Special case for 0 degrees: horizontal motion
		if (angle == 0d || angle == 180d)
		{
			halfX = (int)half;
			halfY = 0;
		}
		
		// Special case for 90 degrees: vertical motion
		else if (angle == 90d)
		{
			halfX = 0;
			halfY = (int)half;
		}
		
		// Other cases, in which cosPhi and sinPhi are surely < 1
		else
		{
			// Absolute value fixes the ceil/floor problem to simulate MATLAB's fix()
			halfX = (int)Math.floor(Math.abs(half * cosPhi + xSign));
			
			// Always positive between 0 and 180 degrees
			halfY = (int)Math.floor(half * sinPhi + 1d); 
		}
		
		// Main algorythm cycle
		for (int yi = 0; yi <= halfY; yi++)
		{
			for (int xi = 0; xi <= halfX; xi++)
			{
				// Define shortest distance from a pixel to the rotated line 
				double x = xi * xSign;
				double y = yi;
				
				// Distance perpendicular to the line
				double distToLine = y * cosPhi - x * sinPhi;
				double radius = Math.sqrt(x * x + y * y);
				
				// Points beyond the line's end-point but within the line width
				if (radius >= half && Math.abs(distToLine) <= 1d)
				{
					// Distance to the line's end-point parallel to the line 
					double xToLastPix = half - Math.abs((x + distToLine * sinPhi) / cosPhi);
     				distToLine = Math.sqrt(distToLine * distToLine + xToLastPix * xToLastPix);
				}
				
				// The more the pixel belongs to the line, the more distToLine is near 1.
				// In fact, if it's exactly 1, the pixel lies on the line.
				distToLine = 1d - Math.abs(distToLine);
				
				// If distToLine is negative, the pixel doesn't lie on the line.
				if (distToLine < 0d) distToLine = 0d;

				// Indexes w/ circular shift
				int i1 = getShiftedIndex(-xi,  xSign * yi);
				int i2 = getShiftedIndex( xi, -xSign * yi);
				
				// Save data
				otf[i1] = distToLine;
				normFactor += distToLine;	
			
				// Avoid redundant writings and erroneous normFactor calculation
				if (i1 != i2)
				{
					otf[i2] = distToLine;
					normFactor += distToLine;
				}
			}
		}
		
		// Normalization cycle
		for (int yi = 0; yi <= halfY; yi++)
		{
			for (int xi = 0; xi <= halfX; xi++)
			{
				// Indexes, w/ circular shift
				int i1 = getShiftedIndex(-xi,  xSign * yi);
				int i2 = getShiftedIndex( xi, -xSign * yi);
				
				// Normalize
				otf[i1] /= normFactor;
				
				// Avoid erroneous normalizations
				if (i1 != i2)
					otf[i2] /= normFactor;
			}	
		}
		
		// Compute FFT in order to get the OTF from the PSF
		fft.realForwardFull(otf);
	}
}
