/**
 * Copyright 2011 Marco Buzzanca. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright notice, this list of
 *       conditions and the following disclaimer.
 * 
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list
 *       of conditions and the following disclaimer in the documentation and/or other materials
 *       provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are those of the
 * authors and should not be interpreted as representing official policies, either expressed
 * or implied, of the copyright holder.
 */

import ij.*;
import ij.io.*;
import ij.process.*;
import ij.gui.*;
import ij.plugin.frame.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.text.*;
import java.util.*;
import java.net.*;
import java.io.*;
import java.security.*;
import javax.imageio.*;
import shx.gui.awt.*;
import shx.web.*;
import shx.security.*;
import shx.math.*;

public class Motion_Deblur extends PlugInFrame 
{
	private static final int SLIDER_MAX = 1000001;
	private static final int SLIDER_VISIBLE = SLIDER_MAX - 1;
	private static final int SLIDER_INC_CLICK = SLIDER_VISIBLE / 100;
	private static final int SLIDER_INC_BLOCK = SLIDER_VISIBLE / 10;
	
	private static final int RED_CHANNEL = 0;
	private static final int GREEN_CHANNEL = 1;
	private static final int BLUE_CHANNEL = 2;
	
	private static final String templateFilename = "ReportTemplate.html";
	
	private NumericTextField ntfShift;
	private NumericTextField ntfAngle;
	private NumericTextField ntfSNR;
	
	private Scrollbar sbShift;
	private Scrollbar sbAngle;
	private Scrollbar sbSNR;
	
	private Checkbox chkFilter;
	
	private ImageWindow srcWindow;
	private ImageWindow destWindow;
	private ImageWindow filterWindow;
	
	private MotionDeblurFilter filter;
	private DecimalFormat df;
	private FileDigest md5;
	
	private double maxShift;
	private double shift;
	private double angle;
	private double snr;
	
	private double destShift;	
	private double destAngle;	
	private double destSNR;

	public Motion_Deblur() 
	{
		super("");
		
		// This filter will work on current image
		srcWindow = WindowManager.getCurrentWindow();

		if (srcWindow == null)
		{
			IJ.noImage();
			return;
		}
		
		// Components and variables which aren't referenced outside 
		// the constructor are created here
		
		Panel mainPanel = new Panel(new BorderLayout(10, 10));
		Panel westPanel = new Panel(new GridLayout(3, 2, 5, 10));
		Panel centerPanel = new Panel(new GridLayout(3, 1, 0, 10));
		Panel southPanel = new Panel(new FlowLayout());
		
		Button btnReport = new Button("Generate Report");
		Button btnApply = new Button("Apply filter");
		
		int width = srcWindow.getImagePlus().getWidth();
		int height = srcWindow.getImagePlus().getHeight();
		
		// Initialize attributes
		
		ntfShift = new NumericTextField("0.00000", 10);
		ntfAngle = new NumericTextField("0.00000", 10);
		ntfSNR = new NumericTextField("0.01000", 10);
		
		sbShift = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, SLIDER_MAX);
		sbShift.setUnitIncrement(SLIDER_INC_CLICK);
		sbShift.setBlockIncrement(SLIDER_INC_BLOCK);
		
		sbAngle = new Scrollbar(Scrollbar.HORIZONTAL, 0, 0, 0, SLIDER_MAX);
		sbAngle.setUnitIncrement(SLIDER_INC_CLICK);
		sbAngle.setBlockIncrement(SLIDER_INC_BLOCK);
		
		sbSNR = new Scrollbar(Scrollbar.HORIZONTAL, 600000, 0, 0, SLIDER_MAX);
		sbSNR.setUnitIncrement(SLIDER_INC_CLICK);
		sbSNR.setBlockIncrement(SLIDER_INC_BLOCK);
		
		chkFilter = new Checkbox("Show OTF");
		
		df = new DecimalFormat("0.00000");
		filter = new MotionDeblurFilter(width, height);
		md5 = new FileDigest("MD5");
		
		maxShift = (width > height ? height / 2 : width / 2);
		
		shift = 0d;
		angle = 0d;
		snr = 0.01d;
		
		// Associate events
		
		btnApply.addActionListener(getBtnFilterListener());
		btnReport.addActionListener(getBtnReportListener());
		
		ntfShift.addActionListener(getNtfShiftListener());
		ntfAngle.addActionListener(getNtfAngleListener());
		ntfSNR.addActionListener(getNtfSNRListener());
		
		sbShift.addAdjustmentListener(getSbShiftListener());
		sbAngle.addAdjustmentListener(getSbAngleListener());
		sbSNR.addAdjustmentListener(getSbSNRListener());
		
		srcWindow.getCanvas().addMouseListener(getMouseListener());
		
		// Add components to panels and pack()

		westPanel.add(new Label("Shift"));
		westPanel.add(ntfShift);
		westPanel.add(new Label("Angle"));
		westPanel.add(ntfAngle);
		westPanel.add(new Label("SNR"));
		westPanel.add(ntfSNR);
		mainPanel.add("West", westPanel);

		centerPanel.add(sbShift);
		centerPanel.add(sbAngle);
		centerPanel.add(sbSNR);
		mainPanel.add("Center", centerPanel);

		southPanel.add(btnApply);
		southPanel.add(btnReport);
		southPanel.add(chkFilter);
		mainPanel.add("South", southPanel);

		add(mainPanel);
		pack();

		// Initialize frame and show
		String srcTitle = srcWindow.getImagePlus().getTitle();
		
		IJ.setTool(Toolbar.LINE);
		GUI.center(this);
		
		setTitle("Motion deblur (" + srcTitle + ")");
		setResizable(false);
		setSize(500, 185);
		setVisible(true);
	}

	@Override
	public Insets getInsets()
	{
		Insets i = super.getInsets();

		i.top += 10;
		i.right += 10;
		i.bottom += 10;
		i.left += 10;

		return i;
	}
	
	private MouseListener getMouseListener()
	{
		return new MouseListener()
		{
			private Point p1;
		
			public void mousePressed(MouseEvent e) 
			{
				if (Toolbar.getToolId() != Toolbar.LINE)
					return;
					
				ImageCanvas canvas = srcWindow.getCanvas();
				
				int width = srcWindow.getImagePlus().getWidth();
				int height = srcWindow.getImagePlus().getHeight();
				
				int x = MathEx.cap(0, width, canvas.offScreenX(e.getX()));
				int y = MathEx.cap(0, height, canvas.offScreenY(e.getY()));
				
				p1 = new Point(x, y);
			}
			
			public void mouseReleased(MouseEvent e) 
			{
				if (Toolbar.getToolId() != Toolbar.LINE)
					return;
					
				ImageCanvas canvas = srcWindow.getCanvas();
				
				int width = srcWindow.getImagePlus().getWidth();
				int height = srcWindow.getImagePlus().getHeight();
				
				int x = MathEx.cap(0, width, canvas.offScreenX(e.getX()));
				int y = MathEx.cap(0, height, canvas.offScreenY(e.getY()));
				
				Point p2 = new Point(x, y);
				
				shift = p1.distance(p2);
				angle = Math.toDegrees(Math.acos(Math.abs(p1.x - p2.x) / shift));
				
				if (shift > maxShift)
					shift = maxShift;
				
				// Angle is NaN if shift is 0
				
				else if (shift == 0d)
					angle = 0d;
				
				ntfShift.setText(df.format(shift));
				ntfAngle.setText(df.format(angle));
				
				sbShift.setValue((int)Math.round(shift / maxShift * SLIDER_VISIBLE));
				sbAngle.setValue((int)Math.round(angle / 180d * SLIDER_VISIBLE));
			}
			
			public void mouseClicked(MouseEvent e) {}
			public void mouseEntered(MouseEvent e) {}
			public void mouseExited(MouseEvent e) {}

		};
	}
	
	private ActionListener getBtnFilterListener()
	{
		return new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				if (srcWindow == null || srcWindow.isClosed())
				{
					IJ.error("Source image window was closed. Please reload the plugin.");
					close();
					return;
				}
				
				int width = srcWindow.getImagePlus().getWidth();
				int height = srcWindow.getImagePlus().getHeight();
				
				ImagePlus srcImage = srcWindow.getImagePlus();
				ImagePlus destImage = new ImagePlus();

				if (shift == 0d)
					destImage.setProcessor(srcImage.getProcessor().duplicate());
				
				else
				{
					filter.calculateMask(shift, angle);
					filter.setSNR(snr);
			
					// Color image
					if (srcImage.getType() == ImagePlus.COLOR_RGB || srcImage.getType() == ImagePlus.COLOR_256)
					{
						ImageProcessor srcProcessor = srcImage.getProcessor().convertToRGB();
						ColorProcessor destProcessor = new ColorProcessor(width, height);
	
						FloatProcessor channel = srcProcessor.toFloat(RED_CHANNEL, null);
						filter.apply((float[])channel.getPixels());
						destProcessor.setPixels(RED_CHANNEL, channel);
	
						channel = srcProcessor.toFloat(GREEN_CHANNEL, channel);
						filter.apply((float[])channel.getPixels());
						destProcessor.setPixels(GREEN_CHANNEL, channel);
	
						channel = srcProcessor.toFloat(BLUE_CHANNEL, channel);
						filter.apply((float[])channel.getPixels());
						destProcessor.setPixels(BLUE_CHANNEL, channel);
	
						destImage.setProcessor(destProcessor);
					}

					// Black & White image
					else 
					{
						ImageProcessor srcProcessor = srcImage.getProcessor().convertToFloat();
						FloatProcessor destProcessor = new FloatProcessor(width, height);
						
						destProcessor.copyBits(srcProcessor, 0, 0, Blitter.COPY);
						filter.apply((float[])destProcessor.getPixels());
	
						destImage.setProcessor(destProcessor);
					}

				}
				
				destImage.changes = true;
				destImage.setTitle(srcImage.getTitle() + " (filtered)");
				
				// Save current shift, angle and snr values. 
				// These values won't change unless this button is pressed again.
				destShift = shift;
				destAngle = angle;
				destSNR = snr;
				
				if (destWindow == null || destWindow.isClosed())
					destWindow = new ImageWindow(destImage);
	
				else 
				{
					destWindow.setImage(destImage);
					destImage.updateAndDraw();
				}
				
				if (shift != 0d && chkFilter.getState())
				{
					double[] otf = filter.getMask();
					float[] pixels = new float[width * height];
					
					// Power spectrum is calculated
					for (int i = 0; i < pixels.length; i++)
						pixels[i] = (float)Math.sqrt(otf[2 * i] * otf[2 * i] + otf[2 * i + 1] * otf[2 * i + 1]);
						
					FloatProcessor filterProcessor = new FloatProcessor(width, height);
					filterProcessor.setPixels(pixels);
					
					ImagePlus filterImage = new ImagePlus();
					filterImage.changes = true;
					filterImage.setTitle("Optical Transfer Function - Power Spectrum");                  
					filterImage.setProcessor(filterProcessor);
					
					if (filterWindow == null || filterWindow.isClosed())
						filterWindow = new ImageWindow(filterImage);
	
					else 
					{
						filterWindow.setImage(filterImage);
						filterImage.updateAndDraw();
					}
				
				}
			}
		};
	}
	
	private ActionListener getBtnReportListener()
	{
		return new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				if (srcWindow == null || srcWindow.isClosed())
				{
					IJ.error("Source image window was closed. Please reload the plugin.");
					close();
					return;
				}
				
				if (destWindow == null || destWindow.isClosed())
				{
					IJ.error("Processed image not available. Apply filter prior to report generation.");
					return;
				}
				
				SaveDialog dialog = new SaveDialog("Save report", IJ.getDirectory("home"), "report", ".html");
				
				// User has pressed cancel
				if (dialog.getFileName() == null)
					return;
				
				HTMLTemplate template = new HTMLTemplate();
				
				// Since plugins are stored in a .jar file, the template file must be accessed as an InputStream
				InputStream templateStream = Motion_Deblur.class.getResourceAsStream(templateFilename);
				
				// Loads from provided stream
				template.load(templateStream);
				
				FileInfo fi = srcWindow.getImagePlus().getOriginalFileInfo();
				String tokenValue;
				
				// Current timestamp
				Calendar cal = Calendar.getInstance();
    			SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
				template.addToken("TIMESTAMP", sdf.format(cal.getTime()));
				
				// User
				template.addToken("USER", System.getProperty("user.name"));
				
				// Workstation
				try { template.addToken("WORKSTATION", InetAddress.getLocalHost().getHostName()); }
				catch (Exception ex) { template.addToken("WORKSTATION", "unknown"); }
				
				// Original file name
				template.addToken("SRC_FILE", fi.fileName);
				
				// Copy original file to report location
				try
				{
					FileInputStream in = new FileInputStream(fi.directory + fi.fileName);
					FileOutputStream out = new FileOutputStream(dialog.getDirectory() + fi.fileName);
					
					byte[] buffer = new byte[1024];
					int i = 0;
					
					while ((i = in.read(buffer)) != -1) 
						out.write(buffer, 0, i);
						
					in.close();
					out.close();
				}
				
				catch (Exception ex) 
				{
					IJ.error("An I/O error occurred.");
					return;
				}
				
				// Hash of original file
				try
				{
					tokenValue = md5.hash(fi.directory + fi.fileName);
					template.addToken("SRC_MD5", tokenValue);
				}
				
				catch (Exception ex)
				{
					IJ.error("Couldn't compute original file's MD5.");
					return;
				}
				
				// Processed file name
				tokenValue = generateFilteredName(fi.fileName);
				template.addToken("DEST_FILE", tokenValue);
				
				// Save processed image data as bmp
				try
				{
					ImagePlus destImage = destWindow.getImagePlus();
					BufferedImage bi = destImage.getBufferedImage();
					File f = new File(dialog.getDirectory() + tokenValue);
					ImageIO.write(bi, "bmp", f);
					
					// Since we are saving the image, let's not ask the user to save the image elsewhere
					destImage.changes = false;
				}
				
				catch (Exception ex) 
				{
					IJ.error("An I/O error occurred.");
					return;
				}
								
				// Hash of destination file			
				try
				{
					tokenValue = md5.hash(dialog.getDirectory() + tokenValue);
					template.addToken("DEST_MD5", tokenValue);
				}
				
				catch (Exception ex)
				{
					IJ.error("Couldn't compute destination file's MD5.");
					return;
				}
				
				// Shift
				template.addToken("SHIFT", df.format(destShift));
				
				// Angle
				template.addToken("ANGLE", df.format(destAngle));
				
				// SNR
				template.addToken("SNR", df.format(destSNR));
				
				// Parses and saves the data
				template.parse();
				template.save(dialog.getDirectory() + dialog.getFileName());
				
				// Opens up a window in your browser ...
				try
				{
					File f = new File(dialog.getDirectory() + dialog.getFileName());
					Desktop.getDesktop().browse(f.toURI());
				}
				
				// .. or shows a message if not supported.
				catch (Exception ex) 
				{
					IJ.showMessage("File \"" + dialog.getFileName() + "\" created successfully!");
				}
			}
		};
	
	}
	
	private ActionListener getNtfShiftListener()
	{
		return new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				try
				{
					shift = Double.parseDouble(ntfShift.getText());
					
					if (shift < 0d)
						throw new Exception();
				}
				
				catch (Exception ex)
				{
					shift = 0d;
					ntfShift.setText("0.00000");
				}
				
				if (shift > maxShift) 
				{
					shift = maxShift;
					ntfShift.setText(df.format(maxShift));
				}

				sbShift.setValue((int)Math.round(shift / maxShift * SLIDER_VISIBLE));
			}
		};
	}
	
	private ActionListener getNtfAngleListener()
	{
		return new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				try
				{
					angle = Double.parseDouble(ntfAngle.getText());
					
					if (angle < 0d)
						throw new Exception();
				}
				
				catch (Exception ex)
				{
					angle = 0d;
					ntfAngle.setText("0.00000");
				}
				
				if (angle > 180d) 
				{
					angle = 180d;
					ntfAngle.setText("180.00000");
				}

				sbAngle.setValue((int)Math.round(angle / 180d * SLIDER_VISIBLE));
			}
		};
	}
	
	private ActionListener getNtfSNRListener()
	{
		return new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				try
				{
					snr = Double.parseDouble(ntfSNR.getText());
					
					if (snr < 0.00001d)
						throw new Exception();
				}
				
				catch (Exception ex)
				{
					snr = 0.00001d;
					ntfSNR.setText("0.00001");
				}
				
				if (snr > 1d) 
				{
					snr = 1d;
					ntfSNR.setText("1.00000");
				}
				
				// Logarithmic representation of noise from 0 (10E-5) to 5 (1.0)
				double logSNR = Math.log10(snr) + 5d;
				
				sbSNR.setValue((int)Math.round(logSNR / 5d * SLIDER_VISIBLE));
			}
		};
	}
	
	private AdjustmentListener getSbShiftListener()
	{
		return new AdjustmentListener()
		{
			public void adjustmentValueChanged(AdjustmentEvent e)
			{		
				shift = e.getValue() * maxShift / SLIDER_VISIBLE;
				ntfShift.setText(df.format(shift));
			}
		};
	}
	
	private AdjustmentListener getSbAngleListener()
	{
		return new AdjustmentListener()
		{
			public void adjustmentValueChanged(AdjustmentEvent e)
			{
				angle = e.getValue() * 180d / SLIDER_VISIBLE;
				ntfAngle.setText(df.format(angle));
			}
		};
	}
	
	private AdjustmentListener getSbSNRListener()
	{
		return new AdjustmentListener()
		{
			public void adjustmentValueChanged(AdjustmentEvent e)
			{
				// Converts the slider range [0 to SLIDER_VISIBLE] into [10E-5 to 1.0]
				snr = Math.pow(10d, e.getValue() * 5d / SLIDER_VISIBLE - 5d);
				ntfSNR.setText(df.format(snr));
			}
		};
	}
	
	// Generates destination filename
	// If source file is named source.xxx, destination file will be named source_filtered.bmp
	private String generateFilteredName(String filename)
	{
		String chunks[] = filename.split("\\.");
		chunks[chunks.length - 2] += "_filtered";
		chunks[chunks.length - 1] = "bmp";
		
		StringBuffer resultBuffer = new StringBuffer();
		
		resultBuffer.append(chunks[0]);
		
		for (int i = 1; i < chunks.length; i++) 
		{
			resultBuffer.append(".");
			resultBuffer.append(chunks[i]);
		}
		
		return resultBuffer.toString();
	}

}
